from abc import ABC, abstractmethod
from multiprocessing import Queue
from queue import Empty
from threading import Event
from typing import Generator

import pika
from pika.adapters.blocking_connection import BlockingChannel, BlockingConnection

from abstra_internals.entities.execution import PreExecution
from abstra_internals.environment import QUEUE_CONCURRENCY, RABBITMQ_EXECUTION_QUEUE
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.producer import QueueMessage
from abstra_internals.utils import deserialize


class Consumer(ABC):
    @abstractmethod
    def done_callback(self, msg: QueueMessage):
        raise NotImplementedError

    @abstractmethod
    def iter(self) -> Generator[QueueMessage, None, None]:
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError


class RabbitConsumer(Consumer):
    channel: BlockingChannel
    connection: BlockingConnection

    def __init__(self, conn_uri: str):
        self.stop_evt = Event()

        self.concurrency = QUEUE_CONCURRENCY
        self.queue = RABBITMQ_EXECUTION_QUEUE

        conn_params = pika.URLParameters(conn_uri)
        self.connection = pika.BlockingConnection(conn_params)

        self.channel = self.connection.channel()
        self.channel.basic_qos(prefetch_count=self.concurrency)
        self.channel.queue_declare(queue=self.queue, durable=True)

    def _callback_threadsafe(self, msg: QueueMessage):
        if self.channel.is_open:
            AbstraLogger.debug(f"[CONSUMER] ack {msg.delivery_tag}")
            self.channel.basic_ack(msg.delivery_tag)
        else:
            AbstraLogger.capture_exception(
                Exception(f"[CONSUMER] cannot ack {msg.delivery_tag}")
            )

    def done_callback(self, msg: QueueMessage):
        self.connection.add_callback_threadsafe(lambda: self._callback_threadsafe(msg))

    def iter(self) -> Generator[QueueMessage, None, None]:
        AbstraLogger.warning("[CONSUMER] starting")
        while True:
            if self.stop_evt.is_set():
                AbstraLogger.warning("[CONSUMER] exiting iteration")
                break

            for method, properties, body in self.channel.consume(
                queue=self.queue, inactivity_timeout=10
            ):
                if not method:
                    break

                yield QueueMessage(
                    preexecution=PreExecution(**deserialize(body.decode("utf-8"))),
                    delivery_tag=method.delivery_tag,
                )

    def stop(self):
        AbstraLogger.warning("[CONSUMER] stoping")
        self.stop_evt.set()

    def close(self):
        AbstraLogger.warning("[CONSUMER] channel canceling")
        self.channel.cancel()
        AbstraLogger.warning("[CONSUMER] channel closing")
        self.channel.close()
        AbstraLogger.warning("[CONSUMER] connection closing")
        self.connection.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.process_data_events(time_limit=10)
        self.close()
        return False

    def __enter__(self) -> "RabbitConsumer":
        return self


# TODO: move out of repositories
class EditorConsumer(Consumer):
    def __init__(self, queue: Queue) -> None:
        self.stop_evt = Event()
        self.queue = queue

    def done_callback(self, msg: QueueMessage):
        pass

    def stop(self):
        self.stop_evt.set()

    def iter(self) -> Generator[QueueMessage, None, None]:
        while True:
            if self.stop_evt.is_set():
                break

            try:
                if msg := self.queue.get(timeout=1):
                    yield msg
            except Empty:
                continue



import webbrowser
from typing import List

from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.version import PackageVersionManager, VersionStatus

RELEASE_NOTES_URL = "https://github.com/abstra-app/abstra-lib/releases"


class OpenChangeLog(LinterFix):
    def __init__(self) -> None:
        self.label = "Open the release notes"

    def fix(self):
        webbrowser.open(RELEASE_NOTES_URL)


class NewVersionOfAbstraAvailableFound(LinterIssue):
    def __init__(self) -> None:
        package_version = PackageVersionManager("abstra")
        self.label = f"Latest version is {package_version.cached_latest_version}, but you have {package_version.current_local_version}."
        self.fixes = [OpenChangeLog()]


class NewVersionOfAbstraAvailable(LinterRule):
    label = "A new version of Abstra Editor is available"
    type = "info"

    def find_issues(self) -> List[LinterIssue]:
        package_version = PackageVersionManager("abstra")
        version_status = package_version.get_version_status()
        is_there_a_new_version = version_status == VersionStatus.OUT_OF_DATE

        if is_there_a_new_version is True:
            return [NewVersionOfAbstraAvailableFound()]
        else:
            return []



from abstra_internals.linter.rules.missing_packages_in_requirements import (
    MissingPackagesInRequirements,
)
from tests.fixtures import BaseTest


class MissingPackagesInRequirementsTest(BaseTest):
    def test_missing_packages_in_requirements_valid_default(self):
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_packages_in_requirements_valid_with_packages(self):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        requirements_file.write_text("pandas==1.0.0")
        script = self.controller.create_script("New script", "script.py")
        code = "import pandas"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_packages_in_requirements_valid_with_packages_using_from_syntax(
        self,
    ):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        requirements_file.write_text("pandas==1.0.0")
        script = self.controller.create_script("New script", "script.py")
        code = "from pandas import foo"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_packages_in_requirements_invalid_requirements_file(self):
        script = self.controller.create_script("New script", "script.py")
        code = "import pandas"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_packages_in_requirements_invalid_requirements_file_using_from_syntax(
        self,
    ):
        script = self.controller.create_script("New script", "script.py")
        code = "from pandas import foo"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_packages_in_requirements_invalid_missing_package(self):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        script = self.controller.create_script("New script", "script.py")
        code = "import pandas"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_packages_with_submodule_import(self):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        script = self.controller.create_script("New script", "script.py")
        code = "import pandas.plotting"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_package_with_submodule_using_from_syntax(self):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        script = self.controller.create_script("New script", "script.py")
        code = "from pandas.plotting import foo"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)



import abstra_internals.email_templates.authentication as authn_template
import abstra_internals.email_templates.message as message_template
import abstra_internals.email_templates.thread_waiting as thread_waiting_template

__all__ = ["authn_template", "thread_waiting_template", "message_template"]



from dataclasses import dataclass

from abstra_internals.environment import (
    CLOUD_API_CLI_URL,
    EDITOR_MODE,
    RABBITMQ_CONNECTION_URI,
    SIDECAR_URL,
)
from abstra_internals.repositories.ai import (
    AiApiHttpClient,
    LocalAiApiHttpClient,
    ProductionAiApiHttpClient,
)
from abstra_internals.repositories.connectors import (
    ConnectorsRepository,
    LocalConnectorsRepository,
    ProductionConnectorsRepository,
)
from abstra_internals.repositories.email import (
    EmailRepository,
    LocalEmailRepository,
    ProductionEmailRepository,
)
from abstra_internals.repositories.execution import (
    ExecutionRepository,
    LocalExecutionRepository,
    ProductionExecutionRepository,
)
from abstra_internals.repositories.execution_logs import (
    ExecutionLogsRepository,
    LocalExecutionLogsRepository,
    ProductionExecutionLogsRepository,
)
from abstra_internals.repositories.jwt_signer import (
    EditorJWTRepository,
    JWTRepository,
    LocalJWTRepository,
    ProductionJWTRepository,
    get_editor_jwt_repository,
)
from abstra_internals.repositories.keyvalue import (
    KVRepository,
    LocalKVRepository,
    ProductionKVRepository,
)
from abstra_internals.repositories.multiprocessing import (
    ForkserverContextRepository,
    MPContextReposity,
    SpawnContextReposity,
)
from abstra_internals.repositories.producer import (
    LocalProducerRepository,
    ProducerRepository,
    ProductionProducerRepository,
)
from abstra_internals.repositories.roles import (
    LocalRolesRepository,
    ProductionRolesRepository,
    RolesRepository,
)
from abstra_internals.repositories.stage_run import (
    LocalStageRunRepository,
    ProductionStageRunRepository,
    StageRunRepository,
)
from abstra_internals.repositories.tables import (
    LocalTablesApiHttpClient,
    ProductionTablesApiHttpClient,
    TablesApiHttpClient,
)
from abstra_internals.repositories.users import (
    LocalUsersRepository,
    ProductionUsersRepository,
    UsersRepository,
)


@dataclass
class Repositories:
    execution_logs: ExecutionLogsRepository
    connectors: ConnectorsRepository
    execution: ExecutionRepository
    mp_context: MPContextReposity
    stage_run: StageRunRepository
    producer: ProducerRepository
    tables: TablesApiHttpClient
    email: EmailRepository
    users: UsersRepository
    roles: RolesRepository
    ai: AiApiHttpClient
    jwt: JWTRepository
    kv: KVRepository
    roles: RolesRepository
    stage_run: StageRunRepository
    tables: TablesApiHttpClient
    users: UsersRepository
    editor_jwt: EditorJWTRepository


def get_editor_repositories():
    mp_context = SpawnContextReposity()

    return Repositories(
        execution=LocalExecutionRepository(mp_context.get_context()),
        stage_run=LocalStageRunRepository(mp_context.get_context()),
        producer=LocalProducerRepository(mp_context.get_context()),
        connectors=LocalConnectorsRepository(CLOUD_API_CLI_URL),
        tables=LocalTablesApiHttpClient(CLOUD_API_CLI_URL),
        email=LocalEmailRepository(CLOUD_API_CLI_URL),
        roles=LocalRolesRepository(CLOUD_API_CLI_URL),
        ai=LocalAiApiHttpClient(CLOUD_API_CLI_URL),
        execution_logs=LocalExecutionLogsRepository(),
        users=LocalUsersRepository(),
        jwt=LocalJWTRepository(),
        kv=LocalKVRepository(),
        editor_jwt=get_editor_jwt_repository(EDITOR_MODE),
        mp_context=mp_context,
    )


def get_prodution_app_repositories():
    if SIDECAR_URL is None or RABBITMQ_CONNECTION_URI is None:
        raise Exception("Production urls are not set")

    return Repositories(
        producer=ProductionProducerRepository(RABBITMQ_CONNECTION_URI),
        execution_logs=ProductionExecutionLogsRepository(SIDECAR_URL),
        connectors=ProductionConnectorsRepository(SIDECAR_URL),
        execution=ProductionExecutionRepository(SIDECAR_URL),
        tables=ProductionTablesApiHttpClient(SIDECAR_URL),
        stage_run=ProductionStageRunRepository(SIDECAR_URL),
        email=ProductionEmailRepository(SIDECAR_URL),
        roles=ProductionRolesRepository(SIDECAR_URL),
        users=ProductionUsersRepository(SIDECAR_URL),
        ai=ProductionAiApiHttpClient(SIDECAR_URL),
        jwt=ProductionJWTRepository(SIDECAR_URL),
        kv=ProductionKVRepository(SIDECAR_URL),
        editor_jwt=get_editor_jwt_repository(EDITOR_MODE),
        mp_context=ForkserverContextRepository(),
    )



import sys
from pathlib import Path
from typing import List

from abstra_internals.linter.linter import LinterFix, LinterIssue, LinterRule
from abstra_internals.settings import Settings


def get_root_and_prefix_path():
    prefix_path = Path(sys.prefix).resolve().as_posix()
    root_path = Settings.root_path.resolve().as_posix()

    return root_path, prefix_path


def running_under_virtualenv() -> bool:
    # pip._internal.utils.virtualenv._running_under_venv
    if sys.prefix != getattr(sys, "base_prefix", sys.prefix):
        return True

    # pip._internal.utils.virtualenv._running_under_legacy_virtualenv
    if hasattr(sys, "real_prefix"):
        return True

    return False


class AddVenvToAbstraIgnore(LinterFix):
    label = "Add virtual env to abstra ignore"

    def fix(self):
        root_path, prefix_path = get_root_and_prefix_path()
        venv_folder = prefix_path.replace(root_path, "").lstrip("/")

        abstraignore_file = Settings.root_path / ".abstraignore"
        with abstraignore_file.open("a") as file:
            file.write("\n")
            file.write(venv_folder)


class VenvInBundleFound(LinterIssue):
    def __init__(self) -> None:
        self.label = "You have not ignored the virtualenv folder"
        self.fixes = [AddVenvToAbstraIgnore()]


class VenvInBundle(LinterRule):
    label = "You can't add virtual env to the bundle"
    type = "bug"

    def virtualenv_inside_project(self) -> bool:
        root_path, prefix_path = get_root_and_prefix_path()
        return prefix_path.startswith(root_path)

    def virtualenv_in_abstraignore(self) -> bool:
        abstraignore_file = Settings.root_path / ".abstraignore"
        if not abstraignore_file.exists():
            return False

        root_path, prefix_path = get_root_and_prefix_path()
        venv_folder = prefix_path.replace(root_path, "").lstrip("/")

        for line in abstraignore_file.read_text(encoding="utf-8").split("\n"):
            if line == venv_folder:
                return True

        return False

    def find_issues(self) -> List[LinterIssue]:
        if not running_under_virtualenv():
            return []

        if not self.virtualenv_inside_project():
            return []

        if self.virtualenv_in_abstraignore():
            return []

        return [VenvInBundleFound()]



import threading
from typing import List

from abstra_internals.controllers.execution_client import ExecutionClient
from abstra_internals.controllers.sdk import ExecutionSDKContext
from abstra_internals.entities.execution import Execution
from abstra_internals.interface.sdk.user_exceptions import ExecutionNotFound
from abstra_internals.repositories.factory import Repositories


# TODO: add context manager
class ExecutionStore:
    rlock = threading.RLock()
    store: List[ExecutionSDKContext] = []

    @classmethod
    def clear(cls):
        data = cls.get_by_thread()
        with cls.rlock:
            cls.store.remove(data)

    @classmethod
    def set(
        cls, execution: Execution, client: ExecutionClient, repositories: Repositories
    ):
        with cls.rlock:
            cls.store.append(
                ExecutionSDKContext(
                    thread_id=threading.get_ident(),
                    repositories=repositories,
                    execution=execution,
                    client=client,
                )
            )

    @classmethod
    def get_by_thread(cls) -> ExecutionSDKContext:
        thread_id = threading.get_ident()
        with cls.rlock:
            for data in cls.store:
                if data.thread_id == thread_id:
                    return data
            raise ExecutionNotFound()

    @classmethod
    def get_by_execution_id(cls, execution_id: str) -> ExecutionSDKContext:
        with cls.rlock:
            for data in cls.store:
                if data.execution.id == execution_id:
                    return data
            raise ExecutionNotFound()



class UnsetThread(Exception):
    def __init__(self):
        self.message = ""
        super().__init__(self.message)


class ThreadAlreadyRunning(Exception):
    def __init__(self):
        self.message = "Thread already running."
        super().__init__(self.message)


class ThreadNotFound(Exception):
    def __init__(self):
        self.message = "Thread not found for the given filter."
        super().__init__(self.message)


class GetUserFailed(Exception):
    def __init__(self):
        self.message = "Failed to get the current user."
        super().__init__(self.message)


class IncompatibleStage(Exception):
    pass


class MissingDependencyError(Exception):
    def __init__(self, package: str):
        self.message = f"The package {package} is required to use this feature. Please install it by running `pip install {package}` and add it to your requirements."
        super().__init__(self.message)


class ExecutionNotFound(Exception):
    pass



from abstra_internals.interface.sdk.files import (
    create_public_link,
    get_persistent_dir,
    get_public_dir,
)

__all__ = [
    "create_public_link",
    "get_persistent_dir",
    "get_public_dir",
]



import abstra.workflows as aw

# Use Scripts to run Python code after any workflow Stage

# You can store data in the current thread,
# so it will be available in the next stages
aw.set_data("script_output", "Some data from the script!")



# flake8: noqa
# Can't use __all__ yet because of the wildcard imports and code generation

from abstra_internals.interface.sdk.forms.generated.inputs import *
from abstra_internals.interface.sdk.forms.generated.outputs import *

from abstra_internals.interface.sdk.forms.list_item_schema import ListItemSchema
from abstra_internals.interface.sdk.forms.page import Page
from abstra_internals.interface.sdk.forms.reactive import reactive
from abstra_internals.interface.sdk.forms.step import run_steps

from abstra_internals.interface.sdk.forms.other import (
    url_params,
    execute_js,
    get_query_params,
    get_user,
    redirect,
)



import io
import pathlib
import shutil
from typing import TYPE_CHECKING, Union

import requests

if TYPE_CHECKING:
    from PIL.Image import Image

from abstra_internals.constants import get_uploads_dir
from abstra_internals.utils.file import get_random_filepath, get_tmp_upload_dir
from abstra_internals.widgets.apis import upload_file

FILE_TYPES = {
    "txt": {
        "mime": "text/plain",
        "extensions": ["txt"],
    },
    "md": {
        "mime": "text/markdown",
        "extensions": ["md"],
    },
    "csv": {
        "mime": "text/csv",
        "extensions": ["csv"],
    },
    "html": {
        "mime": "text/html",
        "extensions": ["html"],
    },
    "css": {
        "mime": "text/css",
        "extensions": ["css"],
    },
    "py": {
        "mime": "text/x-python-script",
        "extensions": ["py"],
    },
    "png": {
        "mime": "image/png",
        "extensions": ["png"],
    },
    "jpg": {
        "mime": "image/jpeg",
        "extensions": ["jpg", "jpeg"],
    },
    "jpeg": {
        "mime": "image/jpeg",
        "extensions": ["jpg", "jpeg"],
    },
    "mp3": {
        "mime": "audio/mp3",
        "extensions": ["mp3"],
    },
    "wav": {
        "mime": "audio/wav",
        "extensions": ["wav"],
    },
    "flac": {
        "mime": "audio/flac",
        "extensions": ["flac"],
    },
    "aac": {
        "mime": "audio/aac",
        "extensions": ["aac"],
    },
    "ogg": {
        "mime": "audio/ogg",
        "extensions": ["ogg"],
    },
    "wma": {
        "mime": "audio/x-ms-wma",
        "extensions": ["wma"],
    },
    "avi": {
        "mime": "video/avi",
        "extensions": ["avi"],
    },
    "mp4": {
        "mime": "video/mp4",
        "extensions": ["mp4"],
    },
    "mkv": {
        "mime": "video/x-matroska",
        "extensions": ["mkv"],
    },
    "mov": {
        "mime": "video/quicktime",
        "extensions": ["mov"],
    },
    "webm": {
        "mime": "video/webm",
        "extensions": ["webm"],
    },
    "flv": {
        "mime": "video/x-flv",
        "extensions": ["flv"],
    },
    "wmv": {
        "mime": "video/x-ms-wmv",
        "extensions": ["wmv"],
    },
    "m4v": {
        "mime": "video/x-m4v",
        "extensions": ["m4v"],
    },
    "zip": {
        "mime": "application/zip",
        "extensions": ["zip"],
    },
    "rar": {
        "mime": "application/vnd.rar",
        "extensions": ["rar"],
    },
    "7z": {
        "mime": "application/x-7z-compressed",
        "extensions": ["7z"],
    },
    "tar": {
        "mime": "application/x-tar",
        "extensions": ["tar"],
    },
    "gzip": {
        "mime": "application/gzip",
        "extensions": ["gzip"],
    },
    "js": {
        "mime": "application/javascript",
        "extensions": ["js"],
    },
    "ts": {
        "mime": "application/typescript",
        "extensions": ["ts"],
    },
    "json": {
        "mime": "application/json",
        "extensions": ["json"],
    },
    "xml": {
        "mime": "application/xml",
        "extensions": ["xml"],
    },
    "yaml": {
        "mime": "application/x-yaml",
        "extensions": ["yaml"],
    },
    "toml": {
        "mime": "application/toml",
        "extensions": ["toml"],
    },
    "pdf": {
        "mime": "application/pdf",
        "extensions": ["pdf"],
    },
    "xls": {
        "mime": "application/vnd.ms-excel",
        "extensions": ["xls"],
    },
    "doc": {
        "mime": "application/msword",
        "extensions": ["doc"],
    },
    "ppt": {
        "mime": "application/vnd.ms-powerpoint",
        "extensions": ["ppt"],
    },
    "docx": {
        "mime": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "extensions": ["docx"],
    },
    "pptx": {
        "mime": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
        "extensions": ["pptx"],
    },
    "image": {
        "mime": "image/*",
        "extensions": ["png", "jpg", "jpeg"],
    },
    "video": {
        "mime": "video/*",
        "extensions": ["avi", "mp4", "mkv", "mov", "webm", "flv", "wmv", "m4v"],
    },
    "audio": {
        "mime": "audio/*",
        "extensions": ["mp3", "wav", "flac", "aac", "ogg", "wma"],
    },
    "text": {
        "mime": "text/*",
        "extensions": ["txt", "md", "csv", "html", "css", "py"],
    },
    "application": {
        "mime": "application/*",
        "extensions": [
            "zip",
            "rar",
            "7z",
            "tar",
            "gzip",
            "js",
            "ts",
            "json",
            "xml",
            "yaml",
            "toml",
            "pdf",
            "xls",
            "doc",
            "ppt",
            "docx",
            "pptx",
        ],
    },
    "unknown": {
        "mime": "*",
        "extensions": [
            "txt",
            "md",
            "csv",
            "html",
            "css",
            "py",
            "png",
            "jpg",
            "jpeg",
            "mp3",
            "wav",
            "flac",
            "aac",
            "ogg",
            "wma",
            "avi",
            "mp4",
            "mkv",
            "mov",
            "webm",
            "flv",
            "wmv",
            "m4v",
            "zip",
            "rar",
            "7z",
            "tar",
            "gzip",
            "js",
            "ts",
            "json",
            "xml",
            "yaml",
            "toml",
            "pdf",
            "xls",
            "doc",
            "ppt",
            "docx",
            "pptx",
            "image",
            "video",
            "audio",
            "text",
            "application",
            "unknown",
            "zip",
            "rar",
            "7z",
            "tar",
            "gzip",
            "js",
            "ts",
            "json",
            "xml",
            "yaml",
            "toml",
            "pdf",
            "xls",
            "doc",
            "ppt",
            "docx",
            "pptx",
        ],
    },
}


def convert_file(file: Union[str, io.IOBase, pathlib.Path, "Image"]) -> str:
    if not file:
        return ""

    if isinstance(file, (io.IOBase, pathlib.Path)):
        return upload_file(file)

    if isinstance(file, str):
        # URL or base64 encoded string
        if file.startswith("http") or file.startswith("data:"):
            return file

        # path to file
        return upload_file(open(file, "rb"))

    from PIL.Image import Image

    if isinstance(file, Image):
        _, file_path = get_random_filepath()
        file.save(str(file_path))
        return upload_file(open(file_path, "rb"))

    # FileResponse. TODO: check with isinstance without circular import
    if hasattr(file, "file"):
        return upload_file(file.file)

    raise ValueError(f"Cannot convert {type(file)}")


def download_to_path(url: str) -> pathlib.Path:
    # circular import
    from abstra_internals.controllers.execution_store import ExecutionStore

    execution_id = ExecutionStore.get_by_thread().execution.id
    save_dir = get_uploads_dir() / execution_id
    save_dir.mkdir(parents=True, exist_ok=True)
    save_path = save_dir / url.split("/")[-1]

    if url.startswith("http://") or url.startswith("https://"):
        with save_path.open("wb") as f, requests.get(url, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

        return save_path

    elif url.startswith("/_files/"):
        tmp_name = url[len("/_files/") :]
        path = get_tmp_upload_dir() / tmp_name
        shutil.copy(path, save_path)

        return save_path

    raise ValueError(f"Cannot download {url}")


def extension_to_mime(extension: str) -> str:
    extension_without_dot = extension.replace(".", "")
    if extension_without_dot in FILE_TYPES:
        return FILE_TYPES[extension_without_dot]["mime"]
    return FILE_TYPES["unknown"]["mime"]



from typing import Dict, List, Tuple, Union

from abstra_internals.controllers.execution_store import ExecutionStore
from abstra_internals.utils import serialize
from abstra_internals.utils.insensitive_dict import CaseInsensitiveDict


def get_raw_request() -> Tuple[str, Dict[str, str], Dict[str, str]]:
    return ExecutionStore.get_by_thread().hook_sdk.get_raw_request()


def get_request() -> Tuple[Union[str, List, Dict], Dict[str, str], Dict[str, str]]:
    return ExecutionStore.get_by_thread().hook_sdk.get_request()


def send_response(body="", status_code=200, headers={}) -> None:
    ExecutionStore.get_by_thread().hook_sdk.set_response(status_code, body, headers)


def send_json(data={}, status_code=200, headers={}):
    if CaseInsensitiveDict(**headers).get("Content-Type") is None:
        headers["Content-Type"] = "application/json"

    send_response(serialize(data), status_code, headers)



from abstra_internals.cloud.server_application import CustomApplication
from abstra_internals.cloud.server_hooks import GunicornOptionsBuilder
from abstra_internals.controllers.main import MainController
from abstra_internals.environment import DEFAULT_PORT
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.factory import get_prodution_app_repositories
from abstra_internals.server.apps import get_cloud_app
from abstra_internals.settings import SettingsController
from abstra_internals.stdio_patcher import StdioPatcher


def run():
    AbstraLogger.init("cloud")
    SettingsController.set_root_path(".")
    SettingsController.set_server_port(DEFAULT_PORT)

    controller = MainController(repositories=get_prodution_app_repositories())
    StdioPatcher.apply(controller)

    options = GunicornOptionsBuilder(controller).build()
    app = get_cloud_app(controller)

    CustomApplication(app, options).run()


if __name__ == "__main__":
    run()



import abc
import datetime
from functools import lru_cache
from typing import Optional

import jwt
import requests

from abstra_internals.environment import PROJECT_ID, SIDECAR_HEADERS
from abstra_internals.jwt_auth import decode_jwt
from abstra_internals.utils import generate_n_digit_code


class JWTRepository(abc.ABC):
    def sign(self, email: str):
        raise NotImplementedError()

    def gen_code(self):
        raise NotImplementedError()

    def sanitize_code(self, code: str) -> str:
        raise NotImplementedError()


class ProductionJWTRepository(JWTRepository):
    base_url: str

    def __init__(self, sidecar_url: str):
        self.base_url = sidecar_url + "/jwt/sign"

    def sign(self, email: str):
        r = requests.post(
            self.base_url, json=dict(email=email), headers=SIDECAR_HEADERS
        )

        if not r.ok:
            raise ValueError(f"Failed to sign JWT: {r.text}")

        return r.json()["jwt"]

    def gen_code(self):
        return generate_n_digit_code(6)

    def sanitize_code(self, code: str) -> str:
        return code


class LocalJWTRepository(JWTRepository):
    def sign(self, email: str):
        return jwt.encode(
            key="fake",
            algorithm="HS256",
            payload={
                "email": email,
                "aud": PROJECT_ID,
                "exp": datetime.datetime.utcnow() + datetime.timedelta(days=7),
            },
        )

    def gen_code(self):
        return "000000"

    def sanitize_code(self, code: str) -> str:
        return "000000"


class EditorJWTRepository(abc.ABC):
    def verify(self, token: Optional[str]) -> bool:
        raise NotImplementedError()

    def decode(self, token: str) -> Optional[dict]:
        raise NotImplementedError()


class LocalEditorJWTRepository(EditorJWTRepository):
    def verify(self, token: Optional[str]) -> bool:
        return True

    def decode(self, token: str) -> Optional[dict]:
        return decode_jwt(token, aud=f"web-editor-{PROJECT_ID}")


class WebEditorJWTRepository(EditorJWTRepository):
    @lru_cache(maxsize=10)
    def __verify(self, token: str) -> bool:
        return self.decode(token) is not None

    def verify(self, token: Optional[str]) -> bool:
        if not token:
            return False
        return self.__verify(token)

    def decode(self, token: str):
        return decode_jwt(token, aud=f"web-editor-{PROJECT_ID}")


def get_editor_jwt_repository(mode) -> EditorJWTRepository:
    if mode == "web":
        return WebEditorJWTRepository()
    return LocalEditorJWTRepository()



import io
import json
import pathlib
from typing import Dict, List, Optional, Union

import pypdfium2 as pdfium

from abstra_internals.repositories.ai import AiApiHttpClient
from abstra_internals.utils.b64 import is_base_64, to_base64
from abstra_internals.utils.string import to_snake_case

Prompt = Union[str, io.IOBase, pathlib.Path]
Format = Dict[str, object]


class AiSDKController:
    def __init__(self, ai_client: AiApiHttpClient):
        self.ai_client = ai_client

    def _build_function_tool_call(self, format: Dict[str, object]) -> Dict[str, object]:
        variable_names = list(format.keys())
        description = "get_" + "_and_".join(variable_names)
        return {
            "name": "get_parameters",
            "description": to_snake_case(description),
            "parameters": {
                "type": "object",
                "properties": format,
                "required": variable_names,
            },
        }

    def _extract_pdf_images(self, file: Prompt) -> List[io.BytesIO]:
        images = []
        for page in pdfium.PdfDocument(file):
            bitmap = page.render(
                scale=1,  # 72dpi resolution
                rotation=0,
            )
            pil_image = bitmap.to_pil()
            image_io = io.BytesIO()
            pil_image.save(image_io, format="png")
            images.append(image_io)
        return images

    def _prompt_is_valid_file(self, prompt: Prompt) -> bool:
        if isinstance(prompt, (io.IOBase, pathlib.Path)) or is_base_64(prompt):
            return True

        try:
            return pathlib.Path(prompt).exists()
        except Exception:
            return False

    def prompt(
        self,
        prompts: List[Prompt],
        instructions: List[str],
        format: Optional[Format],
        temperature: float,
    ):
        messages = []

        for instruction in instructions:
            messages.append(
                {
                    "role": "system",
                    "content": instruction,
                }
            )

        for prompt in prompts:
            if self._prompt_is_valid_file(prompt):
                try:
                    images = self._extract_pdf_images(prompt)
                except pdfium.PdfiumError:
                    # If the file is not a PDF, read it as a single image
                    images = [prompt]
                except FileNotFoundError:
                    raise FileNotFoundError(f"File not found: {prompt}")
                except Exception as e:
                    raise Exception(f"Error reading file: {e}")

                for image in images:
                    base_64_image = to_base64(image)
                    messages.append(
                        {
                            "role": "user",
                            "content": [
                                {
                                    "type": "image_url",
                                    "image_url": {
                                        "url": base_64_image,
                                    },
                                },
                            ],
                        }
                    )
            elif isinstance(prompt, str) and is_base_64(prompt):
                messages.append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": prompt,
                                },
                            },
                        ],
                    }
                )
            elif isinstance(prompt, str):
                messages.append(
                    {
                        "role": "user",
                        "content": prompt,
                    }
                )

        tools = []
        if format:
            function = self._build_function_tool_call(format)
            tools.append({"type": "function", "function": function})

        response = self.ai_client.prompt(messages, tools, temperature)

        if response.get("error"):
            raise Exception(response["error"])

        if format:
            parameters_dict = response["tool_calls"][0]["function"]["arguments"]
            try:
                return json.loads(parameters_dict)
            except json.JSONDecodeError:
                raise Exception(f"Error parsing JSON: {parameters_dict}")

        return response["content"]



import typing


def any_of(cls: typing.Type):
    class AnyOf:
        def __eq__(self, other):
            return isinstance(other, cls)

        def __ne__(self, other):
            return not isinstance(other, cls)

        def __repr__(self) -> str:
            return f"AnyOf({cls.__name__})"

    return AnyOf()



import threading
import time
from typing import Optional

from watchdog.events import FileModifiedEvent, FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from abstra_internals.modules import reload_project_local_modules

IGNORED_FILES = [".abstra/resources.dat", ".abstra/", "abstra.json"]
DEBOUNCE_DELAY = 0.5


class FileChangeEventHandler(FileSystemEventHandler):
    def __init__(self):
        super().__init__()
        self._debounce_timer: Optional[threading.Timer] = None

    def on_modified(self, event: FileSystemEvent):
        has_ignored_paths = any(
            ignored_file in str(event.src_path) for ignored_file in IGNORED_FILES
        )

        if not isinstance(event, FileModifiedEvent) or has_ignored_paths:
            return

        if self._debounce_timer is not None:
            self._debounce_timer.cancel()

        self._debounce_timer = threading.Timer(
            DEBOUNCE_DELAY, self.reload_files_on_change
        )
        self._debounce_timer.start()

    def reload_files_on_change(self):
        try:
            reload_project_local_modules()
        except Exception:
            pass


def watch_file_change_events(path: str):
    event_handler = FileChangeEventHandler()

    observer = Observer()
    observer.schedule(event_handler, path=path, recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()



from datetime import datetime, timezone


def to_utc_iso_string(date: datetime) -> str:
    # datetime.isoformat does not work on python<=3.9 with Z at the end
    return date.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def from_utc_iso_string(isostring: str) -> datetime:
    return datetime.strptime(isostring, "%Y-%m-%dT%H:%M:%S.%fZ").replace(
        tzinfo=timezone.utc
    )


def now_str():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")



import datetime
import re
import time
from typing import Optional
from abstra_internals.widgets.widget_base import Input


class DateInput(Input):
    type = 'date-input'
    empty_value = None

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.value = props.get('initial_value', self.empty_value)
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'hint': self.hint,
            'label': self.label, 'value': self.serialize_value(),
            'required': self.required, 'fullWidth': self.full_width,
            'disabled': self.disabled, 'errors': self.errors}

    def serialize_value(self) ->str:
        if isinstance(self.value, datetime.date):
            date = self.value.isoformat()
            if 'T' in date:
                return date.split('T')[0]
            return date
        if isinstance(self.value, time.struct_time):
            return datetime.datetime.fromtimestamp(time.mktime(self.value)
                ).date().isoformat()
        if isinstance(self.value, str) and re.match('^\\d{4}-\\d{2}-\\d{2}$',
            self.value):
            return self.value
        return ''

    def parse_value(self, value: str) ->Optional[datetime.date]:
        if not value:
            return None
        try:
            split_value = value.split('T')[0].split('-')
            year = int(split_value[0])
            month = int(split_value[1])
            day = int(split_value[2])
            return datetime.date(year, month, day)
        except Exception:
            raise ValueError('Invalid date string format. Expected YYYY-MM-DD.'
                )



import base64
import os
import random
import re
import socket
import sys
import typing as t
from contextlib import closing
from uuid import uuid4

import jsonpath_ng as jp
import simplejson


def generate_n_digit_code(n: int) -> str:
    return str(random.randint(10 ** (n - 1), 10**n - 1))


def serialize(obj, **kwargs):
    return simplejson.dumps(obj, ignore_nan=True, **kwargs)


def deserialize(st) -> t.Any:
    return simplejson.loads(st)


def is_json_serializable(value):
    try:
        serialize(value)
        return True
    except TypeError:
        return False


def is_serializable(st):
    try:
        serialize(st)
        return True
    except Exception:
        return False


def btos(b64):
    return base64.b64decode(b64).decode()


def get_staticmethod(cls, name):
    # hack because python sucks
    method_key = f"_{cls.__name__}{name}"
    return getattr(cls, method_key, None)


def parse_value(cls, value):
    convert = get_staticmethod(cls, "__parse_value")
    return convert(value) if convert else value


def serialize_value(cls, value):
    revert = get_staticmethod(cls, "__serialize_value")
    return revert(value) if revert else value


def random_id(length=10):
    import random
    import string

    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def hash_string(value):
    import hashlib

    sha256_hash = hashlib.sha256()
    sha256_hash.update(value.encode("utf-8"))
    hashed_string = sha256_hash.hexdigest()

    return hashed_string


def get_local_user_id():
    import uuid

    id = str(uuid.getnode())

    return hash_string(id)


def check_is_url(url: str) -> bool:
    return bool(re.match(r"^https?://", url))


def get_free_port(default_port: int) -> int:
    range_start = default_port
    range_end = default_port + 100

    for port in range(range_start, range_end):
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            port_is_available = sock.connect_ex(("localhost", port)) != 0
            if port_is_available:
                return port

    raise Exception(
        f"Could not find a free port in the range {range_start}-{range_end}"
    )


def is_it_true(val: str) -> bool:
    return val.lower() == "true"


def nested_get(data: t.Dict, path: str):
    try:
        expr = jp.parse(path)
        return next(m.value for m in expr.find(data))
    except Exception:
        return None


def is_test_env():
    return "PYTEST_CURRENT_TEST" in os.environ


def is_dev_env() -> bool:
    dev_versions = ["0.0.0"]
    from abstra_internals.utils.packages import get_local_package_version

    return str(get_local_package_version()) in dev_versions


def get_internal_id(obj: object, ensure=True) -> str:
    key = "abstra_uuid"

    if not getattr(obj, key, None) and ensure:
        setattr(obj, key, uuid4().__str__())

    return getattr(obj, key)


def get_local_python_version():
    return f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
